open Temporary
open Ir

  (*************)
  (* registers *)
  (*************)

type register

val string_of_reg: register -> string

val all_registers: register list
    (** list of all physical integer registers *)

val word_registers: register list
    (** integer registers *)

val byte_registers: register list
    (** list of bytes registers (parts of regs in [all_registers]) *)

val lo_of_reg: register -> register
val hi_of_reg: register -> register
    (** correspondance between [all_registers] and [byte_registers] *)

val allocation_registers: unit -> register list
    (** return the list of the registers available for register allocation *)

val caller_save: register list  (** must be saved by the calling function *)
val callee_save: register list  (** must be saved by the called function *)

val arg_registers: register list
    (** registers available for passing args at calls *)

	    (** Some special registers *)
val rv: Temp.temp  (* return value *)
val sl: Temp.temp  (* static link, for function calls *)
val fp: Temp.temp  (* frame pointer *)
val sp: Temp.temp  (* stack pointer *)


val architecture_spec_regs: string -> register list
    (** some architecture specific register, eg:
       x86 : "div", because the division involves particular regs *)

val set_return_value: Ir.exp -> Ir.stm
    (** return the statement which moves the return value at the right place
       no need for a [get_return_value] function, since [Ir.CALL] is an
       expression: the corresponding value is the returned one *)

val temp_of_reg: register -> Temp.temp

val temp_map: (Temp.temp, register) Hashtbl.t
	  (** "colors" of predefined temps (for register allocation)
           *   predefined can be: rv, sl, fp, sp
           * *)



      (***************)
      (*  ACCESSES   *)
      (***************)

type frame
  (** activation record of a function or procedure
   * should contain the name, size, and depth and arguments of the corresponding
   * function
   * *)

val word_size: int
          (** machine word size in bytes *)

type arg_location = Reg | Frame
  (** arguments are passed using registers or the frame 
   * in function call *)

type access
  (** access to a variable: represents how to get it :
    - frame associated
    - in the frame or in a register?
    - pointer to another place?
  *)


exception Not_in_frame

val is_in_frame: access -> bool


val exp_of_sl_access: access -> Ir.exp -> Ir.exp
  (** [exp_of_sl_access a sl] return a tree which accesses the variable of
     access [a] in the frame pointed by the static link [sl] *)

val exp_of_access: access -> Ir.exp
    (** idem, but the access resides in the current frame *)

val follow_sl: frame -> int -> Ir.exp
    (** [follow_sl frame depth] return an IR tree which follows
       the statics links to access to a frame of depth [depth]
       from a frame [frame] *)


      (************)
      (*  FRAME   *)
      (************)

(*  structure of a frame:

   .
   :
   +-----------------+
   | Arg 3           |
   +-----------------+
   | Arg 2           |
   +-----------------+
   | Arg 1           |
   +-----------------+
   | Return Address  | ( <- Caller's SP )
   +=================+
   | Caller's FP     |   <- FP
   +-----------------+
   | Static Link     |
   +-----------------+
   |                 |
   | Local variables |
   |  (spills)       |
   |                 |
   :                 :
   :                 :
   |                 |
   | Args            |   <- SP
   +=================+
   |

*)




val find_frame: Label.label -> frame
	  (** [find_frame l] return the frame associated to name [l] *)

val new_frame_regs:
    ?arglist:arg_location list ->
      Label.label ->
	bool list ->
	  int -> frame
      (** [new_frame_regs argl lab b depth] creates a frame structure for
	 the function labeled [lab], with [b] = true for an
	 arguments which escapes (i.e. is passed by reference)
	 and [depth] is the nested depth of the function.
	 the optional (see [new_frame]) argl list specify how the
	 arguments are passed to the function. If empty, the
	 compiler chooses itself *)

val sizelab: frame -> Label.label
    (** Return the label corresponding to the assembly symbol containing the
       frame size (to use with an "omit-frame-pointer" like option) *)

val depth: frame -> int
  (** [depth f] returns the depth of the function whose frame is [f]
     [0] means global function
     [n+1] means declared in a function of depth [n] *)

val formals: frame -> access list
   (** [formals f] returns the access list of the arguments given to
     the function corresponding to frame [f] *)

val escapes: frame -> bool list
   (** [escape f] returns a list of booleans which are true if their
      corresponding argument of the function of [f] escapes and false
      otherwise *)

val regs_of_args: frame -> Temp.temp list

   (** [regs_of_args f] returns the list of the temporaries corresponding
      to physical registers which contain incoming arguments for [f]
      *)

val name: frame -> string
           (** [name f] return the name of the function corresponding to [f] *)


    (**********)
    (* memory *)
    (**********)

val alloc_local: frame -> bool -> access
	  (** [alloc_local f escape] returns a location for a variable *)

val alloc_in_frame: frame -> int
    (** [alloc_in_frame f] returns an offset from the FP for a memory location*)

val multi_alloc_heap_exp: int -> Ir.exp
(** [multi_alloc_heap_exp n] is used to allocate memory in the heap
  for structures (arrays, records...). [n] is the number of words to
  allocate.
  Since it is not on the stack, it can be kept after the function which created it
  has returned *)


val multi_alloc_heap: access -> int -> Ir.stm
   (** idem, but generate the statement that move the expression returned
      by {!Frame.multi_alloc_heap_exp} at the correct place defined by
      the access *)


    (*********************)
    (*     fragments     *)
    (*********************)

(** a [fragment] is something that will be a part of the final assembly code *)
type fragment =
  | Fun of frame * Ir.stm
  | String of Label.label * string

