(* the Intermediate Representation module *)

type label = Temporary.Label.label
type size = int

type stm =
  | SEQ of stm * stm
	    (* stm1 is executed, then stm2 *)

  | LABEL of label
	    (* target of jumps *)

  | JUMP of exp * label list
	    (* jump to exp, which belongs to the label list *)
	    (*   (useful for later analysis) *)

  | CJUMP of relop * exp * exp * label * label
	    (* if exp1 relop exp2, jump to label1 *)
	    (*               else, jump to label2 *)

  | MOVE of exp * exp
	    (* evaluate exp2 and move it to temp exp1 *)
	    (*              or to memory address exp1 *)
	    (* (exp1 <must> evaluate to TEMP or MEM)  *)

  | EXP of exp

  | NOP

and exp =
  |	BINOP of binop * exp * exp

  |	RELOP of relop * exp * exp

  | MEM of exp
	    (* content of this address in memory *)

  | TEMP of Temporary.Temp.temp
	    (* temporary (in infinite number) *)

  | ESEQ of stm * exp
	    (* execute stm then return the result of exp *)

  | NAME of label

  | CONST of int

  | CALL of exp * exp list
	    (* function or procedure call *)
  | NIL


and binop = PLUS | MINUS | MUL | DIV | MOD
| AND | OR | LSHIFT | RSHIFT
| ARSHIFT | XOR

and relop = EQ | NE | LT | GT | LE | GE
| ULT | ULE | UGT | UGE
		    (* L = lower ;
		       G = greater ;
		       T = strictly ;
		       E = or equal ;
		       U = unsigned
		    *)

let not_relop = function
  | EQ -> NE
  | NE -> EQ
  | LT -> GE
  | GT -> LE
  | LE -> GT
  | GE -> LT
  | ULT -> UGE
  | UGT -> ULE
  | ULE -> UGT
  | UGE -> ULT

let commute_rel = function
  | EQ -> EQ
  | NE -> NE
  | LT -> GT
  | GT -> LT
  | LE -> GE
  | GE -> LE
  | ULT -> UGT
  | UGT -> ULT
  | ULE -> UGE
  | UGE -> ULE


let do_bin_commute = function
  | PLUS | MUL | AND | OR  | XOR -> true
  | MINUS | DIV | MOD | LSHIFT | RSHIFT | ARSHIFT -> false

let step = "  "

let binop = function
  | PLUS -> "+"
  | MINUS -> "-"
  | MUL -> "*"
  | DIV -> "/"
  | MOD -> "%"
  | AND -> "&"
  | OR -> "|"
  | LSHIFT -> "<<"
  | RSHIFT -> ">>"
  | ARSHIFT -> "a>>"
  | XOR -> "xor"

and relop = function
  | EQ -> "="
  | NE -> "<>"
  | LT -> "<"
  | GT -> ">"
  | LE -> "<="
  | GE -> ">="
  | ULT -> "u<"
  | ULE -> "u<="
  | UGT -> "u>"
  | UGE -> "u>="

let bool_to_exp b = if b then CONST 1 else CONST 0



let rec const_exp = function
  | CONST _
  | NAME _ -> true
  | BINOP (b, e, e') ->
      (b = PLUS || b = MINUS || b = MUL) &&   (* TODO: handle more cases *)
      const_exp e && (const_exp e')
  | _ -> false



open Printf

let rec tree_stm indent = function
  | SEQ (stm, stm') ->
      printf "%sSEQ\n" indent ;
      tree_stm (indent ^ step) stm ;
      tree_stm (indent ^ step) stm'

  | LABEL l ->
      printf "%sLabel %s\n" indent (Temporary.Label.name l)

  | JUMP (e, ll) ->
      printf "%sJump [ " indent ;
      List.iter (fun l -> printf "%s ; " (Temporary.Label.name l)) ll;
      printf "]\n" ;
      tree_exp (indent ^ step) e

  | CJUMP (r, e1, e2, l1, l2) ->
      printf "%sCJump '%s'  t=%s  f=%s\n"
	indent
	(relop r)
	(Temporary.Label.name l1)
	(Temporary.Label.name l2) ;
      tree_exp (indent ^ step) e1 ;
      tree_exp (indent ^ step) e2

  | MOVE (e1, e2) ->
      printf "%sMove\n" indent;
      tree_exp (indent ^ step) e1 ;
      tree_exp (indent ^ step) e2

  | EXP e ->
      printf "%sExp\n" indent;
      tree_exp (indent ^ step) e ;

  |	NOP ->
      printf "%sNop\n" indent

and tree_exp indent = function
  | BINOP (b, e1, e2) ->
      printf "%s%s\n" indent (binop b) ;
      tree_exp (indent ^ step) e1 ;
      tree_exp (indent ^ step) e2

  |	RELOP (r, e1, e2) ->
      printf "%s%s\n" indent (relop r) ;
      tree_exp (indent ^ step) e1 ;
      tree_exp (indent ^ step) e2

  | MEM e ->
      printf "%sMem\n" indent;
      tree_exp (indent ^ step) e ;

  | TEMP t ->
      printf "%sTmp %s\n" indent (Temporary.Temp.string t)

  | ESEQ (stm, e) ->
      printf "%sESeq\n" indent ;
      tree_stm (indent ^ step) stm ;
      tree_exp (indent ^ step) e

  | NAME l ->
      printf "%sName %s\n" indent (Temporary.Label.name l)

  | CONST i ->
      printf "%sConst %d\n" indent i

  | CALL (e, el) ->
      printf "%sCall\n" indent ;
      tree_exp (indent ^ step) e ;
      List.iter (tree_exp (indent ^ step)) el

  | NIL -> printf "%sNil\n" indent

let print s = tree_stm "" s ; flush stdout


let rec equal_stm stm1 stm2 = match stm1, stm2 with

    SEQ (s1,s2), SEQ (s1',s2') ->
      equal_stm s1 s1' && equal_stm s2 s2'

  | LABEL l, LABEL l' -> Symbol.equal l l'

  | CJUMP (r,e1,e2,l1,l2), CJUMP (r',e1',e2',l1',l2') ->

      equal_relop r r' && equal_exp e1 e1' && equal_exp e2 e2'
      && Symbol.equal l1 l1' && Symbol.equal l2 l2'

  | JUMP (e1,l1), JUMP (e1',l1') ->

      equal_exp e1 e1' &&

      ( try
	  List.fold_left
	    (fun r -> fun (x,y) -> Symbol.equal x y && r)
	    true
	    (List.combine l1 l1')
	with Invalid_argument _ -> false )

  | MOVE (e1, e2), MOVE (e1', e2') ->
      equal_exp e1 e1' && equal_exp e2 e2'

  | EXP e1, EXP e2 -> equal_exp e1 e2

  | NOP, NOP -> true

  | _ -> false

and equal_exp e1 e2 = match e1, e2 with

    BINOP (b,e1,e2), BINOP (b',e1',e2') ->

      equal_binop b b' && equal_exp e1 e1'
      && equal_exp e2 e2'

  | RELOP (b,e1,e2), RELOP (b',e1',e2') ->

      equal_relop b b' && equal_exp e1 e1'
      && equal_exp e2 e2'

  | MEM e1, MEM e2 -> equal_exp e1 e2

  | TEMP t1, TEMP t2 -> t1 = t2

  | ESEQ (s1,e1), ESEQ (s2,e2) ->
      equal_stm s1 s2 &&
      equal_exp e1 e2

  | NAME l, NAME l' -> Symbol.equal l l'

  | CONST i, CONST j -> i = j

  | CALL (e1,l1), CALL (e2, l2) ->

      equal_exp e1 e2 &&
      ( try
	  List.fold_left
	    (fun r -> fun (x,y) -> equal_exp x y && r)
	    true
	    (List.combine l1 l2)
	with Invalid_argument _ -> false )

  | NIL, NIL -> true

  | _ -> false

and equal_binop b1 b2 = match b1, b2 with

PLUS,PLUS | MINUS,MINUS | MUL,MUL | DIV, DIV | MOD,MOD
| AND,AND | OR,OR | LSHIFT,LSHIFT | RSHIFT,RSHIFT
| ARSHIFT,ARSHIFT | XOR,XOR -> true
| _ -> false

and equal_relop b1 b2 = match b1, b2 with

EQ,EQ | NE,NE | LT,LT | GT,GT | LE,LE | GE,GE
| ULT,ULT | ULE,ULE | UGT,UGT | UGE,UGE -> true
| _ -> false


(******** IR Compaction ********)

let rec opt_s = function

(* particular patterns *)

    EXP (ESEQ (k, NIL)) -> opt_s k

  | EXP (ESEQ (NOP, k)) -> EXP (opt_e k)

  | SEQ (s, EXP NIL) -> SEQ (opt_s s, NOP)

  | EXP NIL -> NOP

  | SEQ (k, SEQ (k',NOP)) -> SEQ (opt_s k, opt_s k')

  | SEQ (k, SEQ (NOP,k')) -> SEQ (opt_s k, opt_s k')

  | SEQ (SEQ (k',NOP), k) -> SEQ (opt_s k', opt_s k)

  | SEQ (SEQ (NOP, k'), k) -> SEQ (opt_s k', opt_s k)

  | EXP (ESEQ (SEQ (k,NOP), ESEQ (k',NIL))) ->
      SEQ (opt_s k, opt_s k')

  | EXP (ESEQ (SEQ (NOP,k), ESEQ (k',NIL))) ->
      SEQ (opt_s k, opt_s k')


(* other cases *)

  | SEQ (s1,s2) -> SEQ (opt_s s1, opt_s s2)

  | JUMP (e, l) -> JUMP (opt_e e, l)

  | CJUMP (r,e1,e2,l1,l2) -> CJUMP (r, opt_e e1, opt_e e2, l1, l2)

  | MOVE (e1, e2) -> MOVE (opt_e e1, opt_e e2)

  | EXP e -> EXP (opt_e e)

  | stm -> stm

and opt_e = function

    BINOP (a,e1,e2) -> BINOP (a, opt_e e1, opt_e e2)

  | RELOP (a,e1,e2) -> RELOP (a, opt_e e1, opt_e e2)

  | MEM e -> MEM (opt_e e)

  | ESEQ (s,e) -> ESEQ (opt_s s, opt_e e)

  | CALL (e,l) -> CALL (opt_e e, List.map opt_e l)

  | e -> e



let rec compact stm =

  let ostm = opt_s stm in

    if equal_stm ostm stm then stm else compact ostm


let to_graph file stm =
  let aux stm =
  Options.verbose 2 (fun () ->
    Printf.eprintf "Printing IR tree in \"%s\"... " file
		    ) ;
  let fd = open_out_bin file in
  let nodes = ref 0 in
  fprintf fd "digraph g {\n\tordering = out ;\n" ;

  let node lab =
    incr nodes ;
    fprintf fd "\tnode%d[label = \"%s\"];\n" !nodes lab ;
    !nodes
  in

  let link x y =
    fprintf fd "\tnode%d -> node%d;\n" x y
  in

  let rec dotty_stm = function
    | SEQ (stm, stm') ->
	let g = dotty_stm stm
	and d = dotty_stm stm' in
	let n = node "Seq"
	in
	link n g ;
	link n d ;
	n

    | LABEL l ->
	node ("Label " ^ (Temporary.Label.name l))

    | JUMP (e, ll) ->
	let g = dotty_exp e in
	let n = node "Jump"
	and d =
	  node
	    (List.fold_left
	       (fun l e -> (Temporary.Label.name e) ^ " ; " ^ l)
	       ""
	       ll
	    ) in
	link n g ;
	link n d ;
	n

    | CJUMP (r, e1, e2, l1, l2) ->
	let g = dotty_exp e1
	and d = dotty_exp e2 in
	let o = node (relop r)
	and i = node (Temporary.Label.name l1)
	and j = node (Temporary.Label.name l2)
	and n = node "CJump" in
	List.iter (link n) [o ; g ; d ; i ; j] ;
	n

    | MOVE (e1, e2) ->
	let g = dotty_exp e1
	and d = dotty_exp e2 in
	let n = node "Move" in
	link n g ;
	link n d ;
	n

    | EXP e ->
	let s = dotty_exp e
	and n = node "Exp" in
	link n s ;
	n

    | NOP ->
	node "Nop"

  and dotty_exp = function
    | BINOP (b, e1, e2) ->
	let g = dotty_exp e1
	and d = dotty_exp e2 in
	let n = node (binop b) in
	link n g ;
	link n d ;
	n

    | RELOP (r, e1, e2) ->
	let g = dotty_exp e1
	and d = dotty_exp e2 in
	let n = node (relop r) in
	link n g ;
	link n d ;
	n

    | MEM e ->
	let s = dotty_exp e
	and n = node "Mem" in
	link n s ;
	n

    | TEMP t ->
	node ("Tmp " ^ Temporary.Temp.string t)

    | ESEQ (stm, e) ->
	let g = dotty_stm stm
	and d = dotty_exp e
	and n = node "ESeq" in
	link n g ;
	link n d ;
	n

    | NAME l ->
	node ("Name " ^ Temporary.Label.name l)

    | CONST i ->
	node ("Const " ^ string_of_int i)

    | CALL (e, el) ->
	let g = dotty_exp e
	and dl =
	  List.map
	    dotty_exp
	    el
	and n = node "Call" in
	List.iter (link n) (g::dl) ;
	n

    | NIL ->
	node ("NIL")

  in
  ignore( dotty_stm stm ) ;

  fprintf fd "}\n" ;
  close_out fd ;
  Options.verbose 2 (fun () ->
    Printf.eprintf "ok\n"
		    )

  in

    if not !Options.nc then
      aux (compact stm)
    else
      aux stm




