(** The Intermediate Representation module
 *)

type label = Temporary.Label.label
type size = int
      (** unused, should be used for defining the size of "thing"
	 but currently considered to be one machine wordsize *)

(*
   if there is two args of the same type of a stm or exp,
   the rule of thumb is to:

   - put the MEM first
   - put the CONST second

   *)


(** The statements of the tree language perform side effects and control flow *)
type stm =
  | SEQ of stm * stm (** [stm1] is executed, then [stm2] *)
  | LABEL of label (** define a label (eg for target of jumps) *)
  | JUMP of exp * label list
	    (** jump to address of [exp], which belongs to the label list
	       (the label list is used for dataflow analysis) *)

  | CJUMP of relop * exp * exp * label * label
	    (** if [exp1 relop exp2], jump to [label1],
	       else jump to [label2] *)

  | MOVE of exp * exp
	    (** evaluate [exp2] and move it to temp [exp1]
	       or to memory address [exp1]
	       ([exp1] {i must} evaluate to [TEMP] or [MEM])  *)

  | EXP of exp (** evaluate [exp] (useful only if it contains side-effects *)
  | NOP (** do nothing *)


(** The expressions stand for the computation of some value
   (possibly with side effects) *)
and exp =
  | BINOP of binop * exp * exp (** compute [e b e'] *)
  | RELOP of relop * exp * exp (** compute boolean operation *)
  | MEM of exp (** content of this address in memory *)
  | TEMP of Temporary.Temp.temp (** temporary (in infinite number) *)
  | ESEQ of stm * exp (** execute [stm] then return the result of [exp] *)
  | NAME of label (** use a label *)
  | CONST of int (** the corresponding integer constant value *)
  | CALL of exp * exp list (** function or procedure call *)
  | NIL (** no value *)


(** Binary operators *)
and binop =
  | PLUS (** addition *)
  | MINUS (** substraction *)
  | MUL (** multiplication *)
  | DIV (** division *)
  | MOD (** modulo *)
  | AND (** logical and *)
  | OR (** logical or *)
  | LSHIFT (** binary left shift *)
  | RSHIFT (** binary right shift *)
  | ARSHIFT (** arithmetic [RSHIFT]: sign inserted instead of 0*)
  | XOR (** logical exclusive or *)

(** Relational operators *)
and relop =
| EQ (** equal *)
| NE (** not equal *)
| LT (** strictly lower *)
| GT (** strictly greater *)
| LE (** lower or equal *)
| GE (** greater or equal *)
| ULT (** unsigned strictly lower *)
| ULE (** unsigned strictly greater *)
| UGT (** unsigned lower or equal *)
| UGE (** unsigned greater or equal *)


val not_relop: relop -> relop
    (** [not_relop r] returns the [relop] [r'] such that
       {[not( x r' y) <=> y r x]} *)

val commute_rel: relop -> relop
    (** [commute_rel r] returns the [relop] [r'] such that
       {[x r y <=> y r' x]} *)

val do_bin_commute: binop -> bool
    (** [do_bin_commute b] returns [true] if {[x b y = y b x]} *)

val const_exp: exp -> bool
    (** Return [true] if the expression contains only [NAME]s, [CONST]s or
       combinations of them *)

val bool_to_exp : bool -> exp
   (** Return a {!Ir.exp} value corresponding to a boolean value *)

val binop : binop -> string
val relop : relop -> string

val print: stm -> unit
    (** [print stm] pretty print the whole tree which root is [stm] on stdout *)


val to_graph: string -> stm -> unit
    (** [to_graph file stm] print the tree of [stm] in [file]. Use dotty
       to display it *)

val equal_stm : stm -> stm -> bool

val equal_exp : exp -> exp -> bool

val equal_binop : binop -> binop -> bool

val equal_relop : relop -> relop -> bool


val compact : stm -> stm
(** simplify a statement tree by removing some nodes.
  Useful for IR pretty printing with dotty *)
